import numpy as np
import warnings
import matplotlib.pylab as plt
from scipy.interpolate import interp1d
import copy
import gaussianprocessderivatives as gp
import om_code.omgenutils as gu
import om_code.omstats as omstats


class fitderiv:
    """
    Smooth and estimate the time derivative of the data via Gaussian processes.

    Summary statistics - the maximal time derivative, the time at which the
    maximal time derivative occurs, the timescale found from inverting the
    maximal time derivative, the maximal value of the smoothed data, and the
    lag time (the time when the tangent from the point with the maximal time
    derivative crosses a line parallel to the time-axis that passes through the
    first data point) - are found and their errors estimated using
    bootstrapping. All statistics can be postfixed by ' err' to find this
    error.

    A summary statistic is given as the median of a distribution of the
    statistic calculated from time series sampled from the optimal Gaussian
    process. Its error is estimated as the interquartile range of this
    distribution.

    After a successful optimisation, the following attributes are generated:

    t: array
        The times specified as input.
    d: array
        The data specified as input.
    f: array
        The mean of the Gaussian process with the optimal hyperparmeters
        at each time point.
    fvar: array
        The variance of the optimal Gaussian process at each time point.
    df: array
        The inferred first time-derivative.
    dfvar: array
        The inferred variance of the first time-derivative.
    ddf: array
        The inferred second time-derivative.
    ddfvar: array
        The inferred variance of the second time-derivative.
    ds: dictionary
        The summary statistics and their estimated errors.

    Examples
    --------
    A typical work flow is:

    >>> from fitderiv import fitderiv
    >>> q= fitderiv(t, od, figs= True)
    >>> q.plotfit('df')

    or potentially

    >>> import matplotlib.pyplot as plt
    >>> plt.plot(q.t, q.d, 'r.', q.t, q.y, 'b')

    Reference
    ---------
    PS Swain, K Stevenson, A Leary, LF Montano-Gutierrez, IBN Clark, J Vogel,
    and T Pilizota.
    Inferring time derivatives including growth rates using Gaussian processes
    Nat Commun 7 (2016) 13766
    """

    def __init__(
        self,
        t,
        d,
        cvfn="matern",
        logs=True,
        noruns=5,
        noinits=100,
        exitearly=False,
        bd=False,
        empirical_errors=False,
        optmethod="L-BFGS-B",
        nosamples=100,
        stats=True,
        statnames=False,
        printstats=True,
        showstaterrors=True,
        warn=False,
        linalgmax=3,
        iskip=False,
    ):
        """
        Smooth data and estimate time derivatives with a Gaussian process.

        Parameters
        ----------
        t: 1D array
            The time points.
        d: array
            The data corresponding to the time points with any replicates given
             as columns.
        cvfn: string
            The type of kernel function for the Gaussian process either 'sqexp'
            (squared exponential) or 'matern' (Matern with nu= 5/2) or 'nn'
            (neural network).
        logs: boolean
            If True, the Gaussian process is used to smooth the natural
            logarithm of the data and the time-derivative is therefore of the
            logarithm of the data.
        noruns: integer, optional
            The number of attempts to be made at optimising the kernel's
            hyperparmeters.
        noinits: integer, optional
            The number of random attempts made to find good initial choices for
            the hyperparameters before running their optimisation.
        exitearly: boolean, optional
            If True, stop at the first successful attempt at optimising the
            hyperparameters otherwise take the best choice from all successful
            optimisations.
        bd: dictionary, optional
            Specifies the limits on the hyperparameters for the Gaussian process.
            For example, bd= {0: [-1, 4], 2: [2, 6]})
            sets confines the first hyperparameter to be between 1e-1 and 1e^4
            and confines the third hyperparmater between 1e2 and 1e6.
        empirical_errors: boolean, optional
            If True, measurement errors are empirically estimated by the
            variance across replicates at each time point.
            If False, the variance of the measurement error is assumed to be
            the same for all time points and its magnitude is a hyperparameter
            that is optimised.
        optmethod: string, optional
            The algorithm used to optimise the hyperparameters, either
            'l_bfgs_b' or 'tnc'.
        nosamples: integer, optional
            The number of bootstrap samples taken to estimate errors in
            statistics.
        stats: boolean, optional
            If True, calculate summary statistics for both the smoothed data and
            the inferred time- derivative.
        statnames: list of strings
            To customise the names of the statistics.
            The default names are:
            'max df' for the maximal time derivative;
            'time of max df' for the time at which the maximal time
            derivative occurs;
            'inverse max df' for the timescale found from inverting the
            maximal time
            derivative;
            'max f' for the maximal value of the smoothed data;
            'lag time' for the lag time defined as the time when the
            tangent from the point with the maximal time derivative
            crosses a line parallel to the time-axis that passes
            through the first data point.
        printstats: boolean, optional
            If True, print the summary statistics calculated.
        showstaterrors: boolean, optional
            If True, display estimated errors for the statistics when printing.
        warn: boolean, optional
            If False, warnings created by covariance matrices that are not
            positive semi-definite are suppressed.
        linalgmax: integer, optional
            The number of times errors generated by underlying linear algebra
            modules during the optimisation by poor choices of the
            hyperparameters should be ignored.
        iskip: integer, optional
            If non-zero, only every iskip'th data point is used to increase
            speed.
        """
        self.linalgmax = linalgmax
        self.success = True
        if not warn:
            # warning generated occasionally when sampling from the Gaussian
            # process likely because of numerical errors
            warnings.simplefilter("ignore", RuntimeWarning)
        try:
            noreps = d.shape[1]
        except IndexError:
            noreps = 1
        self.noreps = noreps
        self.d = np.copy(d)
        self.t = np.copy(t)
        if iskip:
            t = self.t[::iskip]
            d = self.d[::iskip]
        else:
            t = self.t
            d = self.d
        # default bounds for hyperparameters
        bddict = {
            "nn": {0: (-1, 5), 1: (-7, -2), 2: (-6, 2)},
            "sqexp": {0: (-5, 5), 1: (-6, 2), 2: (-5, 2)},
            "matern": {0: (-5, 5), 1: (-4, 4), 2: (-5, 2)},
        }
        # find bounds
        if bd:
            self.bds = gu.mergedicts(original=bddict[cvfn], update=bd)
        else:
            self.bds = bddict[cvfn]
        # log data
        self.logs = logs
        if logs:
            print("Taking natural logarithm of the data.")
            if np.any(np.nonzero(d < 0)):
                print("Warning: Negative data is being set to NaN.")
            # replace zeros and negs so that logs can be applied
            d[np.nonzero(d <= 0)] = np.nan
            # take log of data
            d = np.log(np.asarray(d))
        # run checks and define measurement errors
        if empirical_errors:
            # errors must be empirically estimated
            if noreps > 1:
                lod = [
                    np.count_nonzero(np.isnan(d[:, i])) for i in range(noreps)
                ]
                if np.sum(np.diff(lod)) != 0:
                    print(
                        "The replicates have different number of data "
                        "points, but equal numbers of data points are "
                        "needed for empirically estimating errors."
                    )
                    merrors = None
                else:
                    # estimate errors empirically
                    print("Estimating measurement errors empirically.")
                    merrors = gu.findsmoothvariance(d)
            else:
                print("Not enough replicates to estimate errors empirically.")
                merrors = None
        else:
            merrors = None
        self.merrors = merrors
        # combine data into one array
        tb = np.tile(t, noreps)
        db = np.reshape(d, np.size(d), order="F")
        # check for NaNs
        if np.any(merrors):
            mb = np.tile(merrors, noreps)
            keep = np.intersect1d(
                np.nonzero(~np.isnan(db))[0], np.nonzero(~np.isnan(mb))[0]
            )
        else:
            keep = np.nonzero(~np.isnan(db))[0]
        # remove any NaNs
        da = db[keep]
        ta = tb[keep]
        # check data remains after removing NaNs
        if not np.any(da):
            print("Warning: omfitderiv failed - too many NaNs.")
            self.success = False
        elif np.any(merrors):
            ma = mb[keep]
            if not np.any(ma):
                print("Warning: omfitderiv failed - too many NaNs.")
                self.success = False
        else:
            ma = None
        if self.success:
            self.run(
                cvfn,
                ta,
                da,
                ma,
                noruns,
                noinits,
                exitearly,
                optmethod,
                stats,
                nosamples,
                statnames,
                showstaterrors,
                printstats,
            )

    def run(
        self,
        cvfn,
        ta,
        da,
        ma,
        noruns,
        noinits,
        exitearly,
        optmethod,
        stats,
        nosamples,
        statnames,
        showstaterrors,
        printstats,
    ):
        """Instantiate and run a Gaussian process."""
        try:
            # instantiate GP
            g = getattr(gp, cvfn + "GP")(self.bds, ta, da, merrors=ma)
            print("Using a " + g.description + ".")
            g.info
        except NameError:
            raise Exception("Gaussian process not recognised.")
        # optimise parameters
        g.findhyperparameters(
            noruns,
            noinits=noinits,
            exitearly=exitearly,
            optmethod=optmethod,
            linalgmax=self.linalgmax,
        )
        # display results
        g.results()
        if np.any(ma):
            if len(ma) != len(self.t):
                # NaNs have been removed
                mainterp = interp1d(
                    ta, ma, bounds_error=False, fill_value=(ma[0], ma[-1])
                )
                ma = mainterp(self.t)
        g.predict(self.t, derivs=2, addnoise=True, merrorsnew=ma)
        # results
        self.g = g
        self.logmaxlike = -g.nlml_opt
        self.hparamerr = g.hparamerr
        self.lth = g.lth_opt
        self.f = g.f
        self.df = g.df
        self.ddf = g.ddf
        self.fvar = g.fvar
        self.dfvar = g.dfvar
        self.ddfvar = g.ddfvar
        if stats:
            self.calculatestats(
                nosamples, statnames, showstaterrors, printstats
            )

    def fitderivsample(self, nosamples, newt=None):
        """
        Generate samples from the latent function.

        Both values for the latent function and its first two
        derivatives are returned, as a tuple.

        All derivatives must be sampled because by default all are asked
        to be predicted by the underlying Gaussian process.

        Parameters
        ----------
        nosamples: integer
            The number of samples.
        newt: array, optional
            Time points for which the samples should be made.
            If None, the orginal time points are used.

        Returns
        -------
        samples: a tuple of arrays
            The first element of the tuple gives samples of the latent
            function;
            the second element gives samples of the first time derivative; and
            the third element gives samples of the second time derivative.
        """
        if np.any(newt):
            newt = np.asarray(newt)
            # make prediction for new time points
            gps = copy.deepcopy(self.g)
            gps.predict(newt, derivs=2, addnoise=True)
        else:
            gps = self.g
        samples = gps.sample(nosamples, derivs=2)
        return samples

    def calculatestats(self, nosamples, statnames, showerrors, printstats):
        """
        Calculate statistics from the smoothed data.

        The default names are 'max df', 'time of max df', 'inverse max grad',
        'max f', and 'lag time'.

        Parameters
        ----------
        nosamples: integer
            The number of bootstrap samples used to estimate errors.
        statnames: list of strings, optional
            A list of alternative names for the statistics.
        showerrors: boolean, optional
            If True, display the estimated errors.
        printstats: boolean, optional
            If True, print the summary stats calculated.
        """
        print(f"\nCalculating statistics with {nosamples} samples.")
        if statnames:
            self.stats = statnames
        else:
            self.stats = [
                "min_f",
                "max_f",
                "max_df",
                "time_of_max_df",
                "doubling_time_from_max_df",
                "lag_time",
            ]
        t = self.t
        fs, gs, hs = self.fitderivsample(nosamples)
        # min f
        min_od = fs[np.argmin(fs, 0), np.arange(nosamples)]
        if self.logs:
            min_od = np.exp(min_od)
        # max f
        max_od = fs[np.argmax(fs, 0), np.arange(nosamples)]
        if self.logs:
            max_od = np.exp(max_od)
        # calculate df stats
        im = np.argmax(gs, 0)
        # max df
        mgr = gs[im, np.arange(nosamples)]
        # time of max df
        tmgr = np.array([t[i] for i in im])
        # inverse max df
        dt = np.log(2) / mgr
        # lag time
        lagtime = (
            tmgr
            + (fs[0, np.arange(nosamples)] - fs[im, np.arange(nosamples)])
            / mgr
        )
        # store stats
        ds = {}
        for stname, st in zip(
            self.stats, [min_od, max_od, mgr, tmgr, dt, lagtime]
        ):
            ds[stname] = np.median(st)
            ds[stname + "_err"] = omstats.statserr(st) / 2
        self.ds = ds
        self.nosamples = nosamples
        self.stats2dict(showerrors, printstats)

    def stats2dict(self, showerrors, printstats):
        """
        Create and print a dictionary of statistics.

        The statistics are calculated from the smoothed data and its
        inferred time-derivatives.

        Parameters
        ----------
        showerrors: boolean, optional
            If True, display the errors.
        printstats: boolean optional
            If True, display the statistics.

        Returns
        -------
        statd: dictionary
            The statistics and their errors.
        """
        if showerrors and printstats:
            print("\t(displaying median +/- half interquartile range)\n")
        ds = self.ds
        statd = {}
        lenstr = np.max([len(s) for s in self.stats])
        for s in self.stats:
            statd[s] = ds[s]
            statd[s + "_err"] = ds[s + "_err"]
            if printstats:
                stname = s.rjust(lenstr + 1)
                if showerrors:
                    print(
                        "{:s}= {:6e} +/- {:6e}".format(
                            stname, statd[s], ds[s + "_err"]
                        )
                    )
                else:
                    print("{:s}= {:6e}".format(stname, statd[s]))
        print()
        return statd

    def plotfit(
        self, char="f", errorfac=1, xlabel="time", ylabel=False, figtitle=False
    ):
        """
        Plot the results of the fitting.

        Either the data and the mean of the optimal Gaussian process or
        the inferred time derivatives are plotted.

        Parameters
        ----------
        char: string
            The variable to plot either 'f' or 'df' or 'ddf'.
        errorfac: float, optional
            The size of the errorbars are errorfac times the standard deviation
            of the optimal Gaussian process.
        ylabel: string, optional
            A label for the y-axis.
        figtitle: string, optional
            A title for the figure.
        """
        x = getattr(self, char)
        xv = getattr(self, char + "var")
        if char == "f":
            d = np.log(self.d) if self.logs else self.d
            plt.plot(self.t, d, "r.")
        plt.plot(self.t, x, "b")
        plt.fill_between(
            self.t,
            x - errorfac * np.sqrt(xv),
            x + errorfac * np.sqrt(xv),
            facecolor="blue",
            alpha=0.2,
        )
        if ylabel:
            plt.ylabel(ylabel)
        else:
            plt.ylabel(char)
        plt.xlabel(xlabel)
        if figtitle:
            plt.title(figtitle)



import numpy as np
import om_code.omgenutils as gu


def analyseSunrise(dfd, rcontents, experiment):
    """
    Parse data from an imported Excel file from a Sunrise plate reader.

    Parameters
    --
    dfd: dataframe
        Created by importing the data from a file using Panda's read_excel.
    rcontents: dataframe
        Created by analyseContentsofWells.
    experiment: string
        The name of the experiment.

    Returns
    ------
    rdict: list of dictionaries
        Describes the contents of the plate by experiment, condition, strain,
        time, and well.
    datatypes: list of strings
        Delineates all the types of data in the experiment and is minimally
        ['OD'].
    """
    # extract datatypes
    datatypes = np.array(["OD"], dtype=object)
    # extract times of measurements
    t = (
        gu.rmnans([float(str(ts).split("s")[0]) for ts in dfd.to_numpy()[0]])
        / 3600.0
    )
    # data
    # add to dataframe
    rdict = []
    for index, row in dfd.iterrows():
        if isinstance(row[-1], str) and row[-1][0] in "ABCDEFGH":
            well = row[-1]
            data = row.to_numpy(dtype="float")[:-1]
            if (
                rcontents[well][0] is not None
                and rcontents[well][1] is not None
            ):
                for j in range(len(t)):
                    cons = {
                        "experiment": experiment,
                        "condition": rcontents[well][0],
                        "strain": rcontents[well][1],
                        "time": t[j],
                        "well": well,
                    }
                    dats = {"OD": data[j]}
                    cons.update(dats)
                    rdict.append(cons)
    return rdict, datatypes



# functions to load the data generated by the plate reader
# and to parse the annotation file
import pandas as pd
import numpy as np
import om_code.omerrors as errors
from om_code.analyseTecan import analyseTecan
from om_code.analyseOldTecan import analyseOldTecan
from om_code.analyseSunrise import analyseSunrise


def loadfromplate(
    platereadertype, wdirpath, dname, dsheetnumber, aname, asheetnumber
):
    """Create dataframe from input files generated by the plate reader."""
    experiment = dname.split(".")[0]

    # import and process plate contents file
    allconditions, allstrains, alldata, rcontents = analyseContentsofWells(
        wdirpath, experiment, aname, asheetnumber
    )

    # import data created by plate reader
    try:
        print("loading", dname)
        dfd = pd.read_excel(str(wdirpath / dname), sheet_name=dsheetnumber)
    except FileNotFoundError:
        raise errors.FileNotFound(str(wdirpath / dname))

    # extract data from the plate reader file
    if platereadertype == "Tecan":
        rdict, datatypes = analyseTecan(dfd, rcontents, experiment)
    elif platereadertype == "old Tecan":
        rdict, datatypes = analyseOldTecan(dfd, rcontents, experiment)
    elif platereadertype == "Sunrise":
        rdict, datatypes = analyseSunrise(dfd, rcontents, experiment)
    else:
        raise errors.UnknownPlateReader(platereadertype)
    rdict = pd.DataFrame(rdict)
    datatypes = list(datatypes)
    # return dataframes for raw and processed data, error
    return rdict, allconditions, allstrains, alldata, experiment, datatypes


###


def analyseContentsofWells(wdirpath, experiment, aname, asheetnumber):
    """
    Load and parse ContentsofWells file.

    Returns rcontents, a dictionary with the contents of each well indexed
    by well, and alldata, a dictionary describing the contents of each well
    indexed by experiment.
    """
    import re

    try:
        alldata = {}
        # import contents of the wells
        anno = pd.read_excel(
            str(wdirpath / aname), index_col=0, sheet_name=asheetnumber
        )
        alldata[experiment] = []
        rcontents = {}
        # run through and parse content of each well
        for x in np.arange(1, 13):
            for y in "ABCDEFGH":
                well = y + str(x)
                if (
                    isinstance(anno[x][y], str)
                    and anno[x][y] != "contaminated"
                ):
                    s, c = anno[x][y].split(" in ")
                    # standardise naming of wells with no strains
                    s = re.sub("(null|NULL)", "Null", s)
                    rcontents[well] = [c.strip(), s.strip()]
                    alldata[experiment].append(
                        rcontents[well][1] + " in " + rcontents[well][0]
                    )
                else:
                    rcontents[well] = [None, None]
        # create summary descriptions of the well contents
        alldata[experiment] = list(np.unique(alldata[experiment]))
        allconditions = list(
            np.unique(
                [
                    rcontents[well][0]
                    for well in rcontents
                    if rcontents[well][0] is not None
                ]
            )
        )
        allstrains = list(
            np.unique(
                [
                    rcontents[well][1]
                    for well in rcontents
                    if rcontents[well][0] is not None
                ]
            )
        )
        return allconditions, allstrains, alldata, rcontents
    except FileNotFoundError:
        raise errors.FileNotFound(str(wdirpath / aname))



# Find and analyse mid-log growth
import matplotlib.pylab as plt
import numpy as np
from nunchaku import Nunchaku

import om_code.admin as admin
import om_code.sunder as sunder


def find_midlog_stats(
    self, stat, min_duration, figs, exps, cons, strs, **kwargs
):
    """Find the stat of all variables in the s dataframe for mid-log growth."""
    for e in exps:
        for c in cons:
            for s in strs:
                # get OD data
                res = sunder.extractwells(
                    self.r, self.s, e, c, s, ["time", "OD"]
                )
                t, od = res[0], res[1]
                if len(od):
                    res_dict = {col: [] for col in self.s.columns}
                    print(f"\nFinding mid-log growth for {e} : {s} in {c}")
                    # run nunchaku on log(OD)
                    od[od < 0] = np.finfo(float).eps
                    X = np.median(t, axis=1)
                    Y = np.log(od.T)
                    nc = Nunchaku(
                        X,
                        Y,
                        prior=[-5, 5],
                        **kwargs,
                    )
                    num_segs, evidence = nc.get_number(max_num=6)
                    bds, bds_std = nc.get_iboundaries(num_segs)
                    res_df = nc.get_info(bds)
                    # pick midlog segment
                    t_st, t_en = pick_midlog(res_df, min_duration)
                    # plot nunchaku's results
                    if figs:
                        nc.plot(res_df, hlmax=None)
                        for tv in [t_st, t_en]:
                            iv = np.argmin((X - tv) ** 2)
                            plt.plot(
                                tv,
                                np.median(Y, axis=0)[iv],
                                "ks",
                                markersize=10,
                            )
                        plt.xlabel("time")
                        plt.ylabel("log(OD)")
                        plt.title(f"mid-log growth for {e} : {s} in {c}")
                        plt.show(block=False)
                    # find means over mid-log growth
                    sdf = self.s[
                        (self.s.experiment == e)
                        & (self.s.condition == c)
                        & self.s.strain.str.contains(s)
                    ]
                    midlog_sdf = sdf[(sdf.time >= t_st) & (sdf.time <= t_en)]
                    stat_fn = getattr(midlog_sdf, stat)
                    stats = stat_fn(numeric_only=True)
                    # store results in dict
                    res_dict["experiment"].append(e)
                    res_dict["condition"].append(c)
                    res_dict["strain"].append(s)
                    for key, value in zip(stats.index, stats.values):
                        res_dict[key].append(value)
                    # add "mid-log" to data names
                    res_dict = {
                        (
                            stat + "_midlog_" + k
                            if k not in ["experiment", "condition", "strain"]
                            else k
                        ): v
                        for k, v in res_dict.items()
                    }
                    # add to sc dataframe
                    admin.add_to_sc(self, res_dict)


def pick_midlog(res_df, min_duration):
    """Find midlog from nunchaku's results dataframe."""
    # find duration of each segment
    xs = np.array([np.array(i) for i in res_df["x range"].values])
    res_df["duration"] = np.diff(xs, axis=1)
    # midlog had a minimal duration and positive growth rate
    sdf = res_df[(res_df.duration > min_duration) & (res_df.gradient > 0)]
    # find mid-log growth - maximal specific growth rate
    ibest = sdf.index[sdf.gradient.argmax()]
    t_st, t_en = sdf["x range"][ibest]
    return t_st, t_en



# functions to calculate the statistics
import numpy as np
import pandas as pd
from scipy import integrate
from scipy.interpolate import interp1d
from scipy.signal import find_peaks
from scipy.stats import iqr


def statserr(d, **kwargs):
    """
    Use the half the interquartile range as an error.

    For errors in statistics calculated from samples from a
    Gaussian process to be consistent with fitderiv.
    """
    return iqr(d)


def cleansc(self):
    """Ensure that NaNs do not change numeric variables from being floats."""
    floatvars = [
        "log2_OD_ratio",
        "log2_OD_ratio_err",
        "local_max_gr",
        "local_max_gr_err",
        "time_of_local_max_gr",
        "time_of_local_max_gr_err",
        "area_under_gr_vs_OD",
        "area_under_gr_vs_OD_err",
        "normalised_area_under_gr_vs_OD",
        "normalised_area_under_gr_vs_OD_err",
        "area_under_OD",
        "area_under_OD_err",
        "normalised_area_under_OD",
        "normalised_area_under_OD_err",
        "OD_logmaxlike",
        "max_gr",
        "max_gr_err",
        "time_of_max_gr",
        "time_of_max_gr_err",
        "doubling_time",
        "doubling_time_err",
        "max_OD",
        "max_OD_err",
        "lag_time",
        "lag_time_err",
    ]
    for var in floatvars:
        if var in self.sc.columns:
            self.sc[var] = self.sc[var].astype(float)


def findsummarystats(
    dtype,
    derivname,
    logs,
    nosamples,
    f,
    t,
    e,
    c,
    s,
    findareas,
    figs,
    plotlocalmax,
    axgr,
    showpeakproperties,
    **kwargs,
):
    """Find summary statistics from GP fit to time series of dtype."""
    warning = None
    odtype = "log(" + dtype + ")" if logs else dtype
    outdf = pd.DataFrame(
        {
            "experiment": e,
            "condition": c,
            "strain": s,
            "time": t,
            "f" + odtype: f.f,
            "f" + odtype + "_err": np.sqrt(f.fvar),
            derivname: f.df,
            derivname + "_err": np.sqrt(f.dfvar),
            "d/dt_" + derivname: f.ddf,
            "d/dt_" + derivname + "_err": np.sqrt(f.ddfvar),
        }
    )
    # check derivative has been sensibly defined
    if (
        np.max(np.abs(f.df)) < 1.0e-20
        and np.max(np.abs(np.diff(f.dfvar))) < 1.0e-20
    ):
        warning = (
            "\nWarning: getstats may have failed for "
            + e
            + ": "
            + s
            + " in "
            + c
        )
    # find summary statistics
    fs, gs, hs = f.fitderivsample(nosamples)
    # find local maximal derivative
    da, dt = findlocalmaxderiv(
        f, gs, axgr, figs, plotlocalmax, showpeakproperties, **kwargs
    )
    # find area under df vs f and area under f vs t
    if findareas:
        adff, andff, aft, anft = findareasunder(t, fs, gs, logs)
    else:
        adff, andff, aft, anft = np.nan, np.nan, np.nan, np.nan
    # store results
    statsdict = {
        "experiment": e,
        "condition": c,
        "strain": s,
        "local_max_" + derivname: np.median(da),
        "local_max_" + derivname + "_err": statserr(da),
        "time_of_local_max_" + derivname: np.median(dt),
        "time_of_local_max_" + derivname + "_err": statserr(dt),
        "area_under_" + derivname + "_vs_" + dtype: np.median(adff),
        "area_under_" + derivname + "_vs_" + dtype + "_err": statserr(adff),
        "normalised_area_under_"
        + derivname
        + "_vs_"
        + dtype: np.median(andff),
        "normalised_area_under_"
        + derivname
        + "_vs_"
        + dtype
        + "_err": statserr(andff),
        "area_under_" + dtype: np.median(aft),
        "area_under_" + dtype + "_err": statserr(aft),
        "normalised_area_under_" + dtype: np.median(anft),
        "normalised_area_under_" + dtype + "_err": statserr(anft),
    }
    return outdf, statsdict, warning


def findlocalmaxderiv(
    f, gs, axgr, figs, plotlocalmax, showpeakproperties, **kwargs
):
    """
    Find the greatest local maxima in the derivative.

    Check the derivative for local maxima and so find the local maximum
    with the highest derivative using samples, gs, of df.

    The keyword variables kwargs are passed to scipy's find_peaks.
    """
    # find peaks in mean df
    lpksmn, lpksmndict = find_peaks(f.df, **kwargs)
    if np.any(lpksmn):
        if showpeakproperties:
            # display properties of peaks
            print("Peak properties\n---")
            for prop in lpksmndict:
                print("{:15s}".format(prop), lpksmndict[prop])
        # da: samples of local max df
        # dt: samples of time of local max df
        da, dt = [], []
        # find peaks of sampled df
        for gsample in np.transpose(gs):
            tpks = find_peaks(gsample, **kwargs)[0]
            if np.any(tpks):
                da.append(np.max(gsample[tpks]))
                dt.append(f.t[tpks[np.argmax(gsample[tpks])]])
        if figs and plotlocalmax:
            # plot local max df as a point
            axgr.plot(
                np.median(dt),
                np.median(da),
                "o",
                color="yellow",
                markeredgecolor="k",
            )
        return da, dt
    else:
        # mean df has no peaks
        return np.nan, np.nan


def findareasunder(t, fs, gs, logs):
    """
    Find areas.

    Given samples of f, as arguments fs, and of df, as arguments gs,
    either find the area under df/dt vs f and the area under f vs t
    or if logs find the area under d/dt log(f) vs f and the area
    under f vs t.
    """
    adff, andff, aft, anft = [], [], [], []
    for fsample, gsample in zip(np.transpose(fs), np.transpose(gs)):
        sf = np.exp(fsample) if logs else fsample
        # area under df vs f: integrand has f as x and df as y
        def integrand(x):
            return interp1d(sf, gsample)(x)

        iresult = integrate.quad(
            integrand, np.min(sf), np.max(sf), limit=100, full_output=1
        )[0]
        adff.append(iresult)
        andff.append(iresult / (np.max(sf) - np.min(sf)))
        # area under f vs t: integrand has t as x and f as y
        def integrand(x):
            return interp1d(t, sf)(x)

        iresult = integrate.quad(
            integrand, np.min(t), np.max(t), limit=100, full_output=1
        )[0]
        aft.append(iresult)
        anft.append(iresult / (np.max(t) - np.min(t)))
    return adff, andff, aft, anft



# functions for taking subsets of the data
import numpy as np

import om_code.omerrors as errors
import om_code.omgenutils as gu


def getsubset(
    self,
    type,
    set="all",
    includes=False,
    excludes=False,
    nonull=False,
    nomedia=False,
):
    """
    Return a subset of either the conditions or strains.

    Parameters
    --
    self: instance of platereader
    type: string
        Either 'c' (conditions) or 's' (strains).
    set: list of strings
        List of items to include (default is 'all').
    includes: string
        Select only items with this string in their name.
    excludes: string
        Ignore any items with this string in their name.
    nonull: boolean
        If True, ignore Null strain.
    nomedia: boolean
        If True, ignores 'media' condition.

    Returns
    -------
    sset: list of strings
    """
    if set == "all" or includes or excludes:
        if type == "c":
            sset = list(
                np.unique(
                    [
                        con
                        for e in self.allconditions
                        for con in self.allconditions[e]
                    ]
                )
            )
            if nomedia and "media" in sset:
                sset.pop(sset.index("media"))
        elif type == "s":
            sset = list(
                np.unique(
                    [
                        str
                        for e in self.allstrains
                        for str in self.allstrains[e]
                    ]
                )
            )
            if nonull and "Null" in sset:
                sset.pop(sset.index("Null"))
        else:
            sset = self.allexperiments
        # find those items containing keywords given in 'includes'
        if includes:
            includes = gu.makelist(includes)
            newset = []
            for s in sset:
                gotone = 0
                for item in includes:
                    if item in s:
                        gotone += 1
                if gotone == len(includes):
                    newset.append(s)
            sset = newset
        # remove any items containing keywords given in 'excludes'
        if excludes:
            excludes = gu.makelist(excludes)
            exs = []
            for s in sset:
                for item in excludes:
                    if item in s:
                        exs.append(s)
                        break
            for ex in exs:
                sset.pop(sset.index(ex))
    else:
        sset = gu.makelist(set)
    if sset:
        # sort by numeric values in list entries
        return sorted(sset, key=gu.natural_keys)
    else:
        if includes:
            raise errors.getsubset(
                "Nothing found for " + " and ".join(includes)
            )
        else:
            raise errors.getsubset("Nothing found")


def getexps(self, experiments, experimentincludes, experimentexcludes):
    """Return list of experiments."""
    if experimentincludes or experimentexcludes:
        exps = getsubset(
            self,
            "e",
            includes=experimentincludes,
            excludes=experimentexcludes,
        )
    elif experiments == "all":
        exps = self.allexperiments
    else:
        exps = gu.makelist(experiments)
    return exps


def getcons(self, conditions, conditionincludes, conditionexcludes, nomedia):
    """Return list of conditions."""
    if conditionincludes or conditionexcludes:
        cons = getsubset(
            self,
            "c",
            includes=conditionincludes,
            excludes=conditionexcludes,
            nomedia=nomedia,
        )
    elif conditions == "all":
        cons = list(
            np.unique(
                [
                    con
                    for e in self.allconditions
                    for con in self.allconditions[e]
                ]
            )
        )
        if nomedia and "media" in cons:
            cons.pop(cons.index("media"))
    else:
        cons = gu.makelist(conditions)
    return sorted(cons, key=gu.natural_keys)


def getstrs(self, strains, strainincludes, strainexcludes, nonull):
    """Return list of strains."""
    if strainincludes or strainexcludes:
        strs = getsubset(
            self,
            "s",
            includes=strainincludes,
            excludes=strainexcludes,
            nonull=nonull,
        )
    elif strains == "all":
        strs = list(
            np.unique(
                [str for e in self.allstrains for str in self.allstrains[e]]
            )
        )
        if nonull and "Null" in strs:
            strs.pop(strs.index("Null"))
    else:
        strs = gu.makelist(strains)
    if nonull and "Null" in strs:
        strs.pop(strs.index("Null"))
    return sorted(strs, key=gu.natural_keys)


def getall(
    self,
    experiments,
    experimentincludes,
    experimentexcludes,
    conditions,
    conditionincludes,
    conditionexcludes,
    strains,
    strainincludes,
    strainexcludes,
    nonull=True,
    nomedia=True,
):
    """Return experiments, conditions, and strains."""
    exps = getexps(self, experiments, experimentincludes, experimentexcludes)
    cons = getcons(
        self, conditions, conditionincludes, conditionexcludes, nomedia
    )
    strs = getstrs(self, strains, strainincludes, strainexcludes, nonull)
    return exps, cons, strs


def extractwells(r_df, s_df, experiment, condition, strain, datatypes):
    """
    Extract a list of data matrices from the r dataframe.

    Each column in each matrix contains the data
    from one well.

    Data is returned for each dtype in datatypes, which may include "time", for
    the given experiment, condition, and strain.
    """
    datatypes = gu.makelist(datatypes)
    # restrict time if necessary
    lrdf = r_df[
        (r_df.time >= s_df.time.min()) & (r_df.time <= s_df.time.max())
    ]
    # extract data
    df = lrdf.query(
        "experiment == @experiment and condition == @condition "
        "and strain == @strain"
    )
    matrices = []
    for dtype in datatypes:
        df2 = df[[dtype, "well"]]
        df2well = df2.groupby("well", group_keys=True)[dtype].apply(list)
        data = np.transpose([df2well[w] for w in df2well.index])
        matrices.append(data)
    if len(datatypes) == 1:
        # return array
        return matrices[0]
    else:
        # return list of arrays
        return matrices



# functions for correcting for non-linearity in the OD, for
# the fluorescence of the media, and for autofluorescence
import importlib.resources as import_files
import re

import gaussianprocessderivatives as gp
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from nunchaku import Nunchaku
from scipy.interpolate import interp1d
from scipy.optimize import minimize_scalar
from statsmodels.nonparametric.smoothers_lowess import lowess

import om_code.omerrors as errors
import om_code.omgenutils as gu
import om_code.sunder as sunder
from om_code.omfitderiv import fitderiv


def _read_dilution_data(fname):
    d = import_files.read_text("om_code", fname)
    res = np.array(re.split(r"\n|\t", d)[:-1]).astype(float)
    if fname == "dilution_data_xiao.tsv":
        od, dilfac = res[::2], res[1::2]
        # missing replicate - use mean of existing ones
        dilfac = np.insert(dilfac, 0, dilfac[0])
        dilfac = dilfac.reshape(int(dilfac.size / 3), 3)
        od = np.insert(od, 0, np.mean(od[:2]))
        od = od.reshape(int(od.size / 3), 3)
    else:
        raise SystemExit("Dilution data unrecognised.")
    return od, dilfac


def findODcorrection(wdirpath, ODfname, figs, bd, gp_results):
    """
    Determine a function to correct OD.

    Use a Gaussian process to fit serial dilution data to correct for
    non-linearities in the relationship between OD and cell density.

    The data are either loaded from file ODfname or the default
    data for haploid yeast growing in glucose are used.
    """
    print("Fitting dilution data for OD correction for non-linearities.")
    if ODfname is not None:
        try:
            od_df = pd.read_csv(
                str(wdirpath / ODfname), sep=None, engine="python", header=None
            )
            print(f"Using {ODfname}.")
            od_data = od_df.to_numpy()
            od, dilfac = od_data[:, 0], od_data[:, 1]
        except (FileNotFoundError, OSError):
            raise errors.FileNotFound(str(wdirpath / ODfname))
    else:
        print("Using default data.")
        fname = "dilution_data_xiao.tsv"
        od, dilfac = _read_dilution_data(fname)
    # run nunchaku
    X = np.mean(dilfac, 1)
    nc = Nunchaku(X, od.T, estimate_err=True, prior=[-5, 5])
    num_regions, _ = nc.get_number(4)
    bds, _ = nc.get_iboundaries(num_regions)
    # linear region starts from origin
    odmatch = np.mean(od, 1)[bds[0]]
    dilfacmatch = X[bds[0]]
    # process data
    dilfac = dilfac.flatten()[np.argsort(od.flatten())]
    od = np.sort(od.flatten())
    # rescale so that OD and dilfac match
    y = dilfac / dilfacmatch * odmatch
    # set up Gaussian process
    bds = {0: (-4, 4), 1: (-4, 4), 2: (-3, 1)}
    # find bounds
    if bd is not None:
        bds = gu.mergedicts(original=bds, update=bd)
    gc = gp.maternGP(bds, od, y)
    # run Gaussian process
    gc.findhyperparameters(noruns=5, exitearly=True, quiet=True)
    if gp_results:
        gc.results()
    gc.predict(od)
    if figs:
        plt.figure()
        gc.sketch(".")
        plt.plot(odmatch, odmatch, "bs")
        plt.grid(True)
        plt.xlabel("OD")
        plt.ylabel("corrected OD (relative cell numbers)")
        if ODfname:
            plt.title("Fitting " + ODfname)
        else:
            plt.title("for haploid budding yeast in glucose")
        plt.show(block=False)
    return gc, odmatch


def performmediacorrection(
    r_df, dtype, exp, condition, figs, commonmedia, frac
):
    """
    Correct data of type dtype for any signal from the media.

    Use lowess to smooth over time the media data from the Null
    wells and subtract the smoothed values from the data.
    """
    # find data for correction with condition equal to commonmedia
    df = r_df.query(
        "experiment == @exp and condition == @commonmedia"
        " and strain == 'Null'"
    )
    if df.empty:
        # no data
        print(
            f' No well annotated "Null" was found for {commonmedia}'
            f"in experiment {exp}."
        )
        print(
            f"-> Correcting for media for {dtype} in {commonmedia} abandoned!"
        )
        return False, None
    else:
        # there is data - change r dataframe
        rtest = (r_df.experiment == exp) & (r_df.condition == condition)
        t, data = df["time"].to_numpy(), df[dtype].to_numpy()
        # find correction
        res = lowess(data, t, frac=frac)
        correctionfn = interp1d(
            res[:, 0],
            res[:, 1],
            fill_value=(res[0, 1], res[-1, 1]),
            bounds_error=False,
        )
        if figs:
            plt.figure()
            plt.plot(t, data, "ro", res[:, 0], res[:, 1], "b-")
            plt.xlabel("time (hours)")
            plt.title(
                exp + ": media correction for " + dtype + " in " + condition
            )
            plt.show(block=False)
        # perform correction
        r_df.loc[rtest, dtype] = r_df[rtest][dtype] - correctionfn(
            r_df[rtest]["time"]
        )
        # check for any negative values
        negvalues = ""
        for s in np.unique(r_df[rtest]["strain"][r_df[rtest][dtype] < 0]):
            if s != "Null":
                wstr = (
                    "\t"
                    + dtype
                    + ": "
                    + s
                    + " in "
                    + condition
                    + " for wells "
                )
                for well in np.unique(
                    r_df[rtest][r_df[rtest].strain == s]["well"][
                        r_df[rtest][dtype] < 0
                    ]
                ):
                    wstr += well + " "
                wstr += "\n"
                negvalues += wstr
        return True, negvalues


def correctauto1(
    self,
    f,
    refstrain,
    figs,
    useGPs,
    flcvfn,
    bd,
    nosamples,
    experiments,
    experimentincludes,
    experimentexcludes,
    conditions,
    conditionincludes,
    conditionexcludes,
    strains,
    strainincludes,
    strainexcludes,
    **kwargs,
):
    """
    Correct autofluorescence for measurements with emissions at one wavelength.

    Corrects for autofluorescence for data with emissions measured at one
    wavelength using the fluorescence of the reference strain
    interpolated to the OD of the tagged strain.

    This method in principle corrects too for the fluorescence of the medium,
    although running correctmedia is still recommended.
    """
    print("Using one fluorescence wavelength.")
    print(f"Correcting autofluorescence using {f[0]}.")
    for e in sunder.getexps(
        self, experiments, experimentincludes, experimentexcludes
    ):
        for c in sunder.getcons(
            self,
            conditions,
            conditionincludes,
            conditionexcludes,
            nomedia=True,
        ):
            # process reference strain
            refstrfn = processref1(self, f, refstrain, figs, e, c)
            # correct strains
            for s in sunder.getstrs(
                self, strains, strainincludes, strainexcludes, nonull=True
            ):
                if not self.sc[
                    (self.sc.experiment == e)
                    & (self.sc.condition == c)
                    & (self.sc.strain == s)
                ][f[0] + "_corrected_for_autofluorescence"].any():
                    od, rawfl = sunder.extractwells(
                        self.r, self.s, e, c, s, ["OD", f[0]]
                    )
                    # no data
                    if od.size == 0 or rawfl.size == 0:
                        print(f"\n-> No data found for {e}: {s} in {c}.\n")
                        continue
                    # correct autofluorescence for each replicate
                    fl = np.transpose(
                        [
                            rawfl[:, i] - refstrfn(od[:, i])
                            for i in range(od.shape[1])
                        ]
                    )
                    fl[fl < 0] = np.nan
                    if useGPs:
                        # get time
                        t = self.s.query(
                            "experiment == @e and condition == @c and strain == @s"
                        )["time"].to_numpy()
                        flperod = samplewithGPs(
                            self,
                            t,
                            fl,
                            od,
                            flcvfn,
                            bd,
                            nosamples,
                            e,
                            c,
                            s,
                            figs,
                            **kwargs,
                        )
                    else:
                        # use only the replicates
                        flperod = np.transpose(
                            [fl[:, i] / od[:, i] for i in range(od.shape[1])]
                        )
                        flperod[flperod < 0] = np.nan
                    # check number of NaN
                    nonans = np.count_nonzero(np.isnan(fl))
                    if np.any(nonans):
                        if nonans == fl.size:
                            print(
                                f"\n-> Corrected fluorescence is all NaN for {e}: {s} in {c}.\n"
                            )
                        else:
                            print(
                                f"Warning - {e}: {s} in {c}\n"
                                f"{nonans} corrected data points are"
                                " NaN because the corrected fluorescence"
                                " was negative.",
                            )
                    # store results
                    bname = "c-" + f[0]
                    autofdict = {
                        "experiment": e,
                        "condition": c,
                        "strain": s,
                        "time": self.s.query(
                            "experiment == @e and condition == @c "
                            "and strain == @s"
                        )["time"].to_numpy(),
                        bname: np.nanmean(fl, 1),
                        bname + "_err": nanstdzeros2nan(fl, 1),
                        bname + "perOD": np.nanmean(flperod, 1),
                        bname + "perOD_err": nanstdzeros2nan(flperod, 1),
                    }
                    addtodataframes(self, autofdict, bname)
                    addrecord(self, f[0], e, c, s)


def processref1(self, f, refstrain, figs, experiment, condition):
    """
    Process reference strain for data with one fluorescence measurement.

    Use lowess to smooth the fluorescence of the reference
    strain as a function of OD.

    Parameters
    ----------
    f: string
        The fluorescence to be corrected. For example, ['mCherry'].
    refstrain: string
        The reference strain. For example, 'WT'.
    figs: boolean
        If True, display fits of the reference strain's fluorescence.
    experiment: string
        The experiment to be corrected.
    condition: string
        The condition to be corrected.

    Returns
    -------
    refstrfn: function
        The reference strain's fluorescence as a function of OD.
    """
    e, c = experiment, condition
    print(f"{e}: Processing reference strain {refstrain} for {f[0]} in {c}.")
    od, fl = sunder.extractwells(self.r, self.s, e, c, refstrain, ["OD", f[0]])
    if od.size == 0 or fl.size == 0:
        raise errors.CorrectAuto(f"{e}: {refstrain} not found in {c}.")
    else:
        odf = od.flatten("F")
        flf = fl.flatten("F")
        # smooth fluorescence as a function of OD using lowess to minimize
        # refstrain's autofluorescence

        def choosefrac(frac):
            res = lowess(flf, odf, frac=frac)
            refstrfn = interp1d(
                res[:, 0],
                res[:, 1],
                fill_value=(res[0, 1], res[-1, 1]),
                bounds_error=False,
            )
            # max gives smoother fits than mean
            return np.max(np.abs(flf - refstrfn(odf)))

        res = minimize_scalar(choosefrac, bounds=(0.1, 0.99), method="bounded")
        # choose the optimum frac
        frac = res.x if res.success else 0.33
        res = lowess(flf, odf, frac=frac)
        refstrfn = interp1d(
            res[:, 0],
            res[:, 1],
            fill_value=(res[0, 1], res[-1, 1]),
            bounds_error=False,
        )
        if figs:
            # plot fit
            plt.figure()
            plt.plot(odf, flf, ".", alpha=0.5)
            plt.plot(res[:, 0], res[:, 1])
            plt.xlabel("OD")
            plt.ylabel(f[0])
            plt.title(e + ": " + refstrain + " for " + c)
            plt.show(block=False)
        return refstrfn


def correctauto2(
    self,
    f,
    refstrain,
    figs,
    useGPs,
    flcvfn,
    bd,
    nosamples,
    experiments,
    experimentincludes,
    experimentexcludes,
    conditions,
    conditionincludes,
    conditionexcludes,
    strains,
    strainincludes,
    strainexcludes,
    **kwargs,
):
    """
    Correct autofluorescence for measurements with two emission wavelengths.

    Corrects for autofluorescence using spectral unmixing for data with
    measured emissions at two wavelengths.

    References
    ----------
    CA Lichten, R White, IB Clark, PS Swain (2014). Unmixing of fluorescence
    spectra to resolve quantitative time-series measurements of gene
    expression in plate readers.
    BMC Biotech, 14, 1-11.
    """
    # correct for autofluorescence
    print("Using two fluorescence wavelengths.")
    print(f"Correcting autofluorescence using {f[0]} and {f[1]}.")
    for e in sunder.getexps(
        self, experiments, experimentincludes, experimentexcludes
    ):
        for c in sunder.getcons(
            self,
            conditions,
            conditionincludes,
            conditionexcludes,
            nomedia=True,
        ):
            # process reference strain
            refqrfn = processref2(self, f, refstrain, figs, e, c)
            # process other strains
            for s in sunder.getstrs(
                self, strains, strainincludes, strainexcludes, nonull=True
            ):
                if s != refstrain and not (
                    self.sc[
                        (self.sc.experiment == e)
                        & (self.sc.condition == c)
                        & (self.sc.strain == s)
                    ][f[0] + "_corrected_for_autofluorescence"].any()
                ):
                    f0, f1, od = sunder.extractwells(
                        self.r, self.s, e, c, s, f.copy() + ["OD"]
                    )
                    if f0.size == 0 or f1.size == 0:
                        print(f"Warning: No data found for {e}: {s} in {c} !!")
                        continue
                    nodata, nr = f0.shape
                    # set negative values to NaNs
                    f0[f0 < 0] = np.nan
                    f1[f1 < 0] = np.nan
                    # use mean OD for correction
                    odmean = self.s.query(
                        "experiment == @e and condition == @c and strain == @s"
                    )["OD_mean"].to_numpy()
                    # correct autofluorescence
                    ra = refqrfn(odmean)
                    fl = applyautoflcorrection(self, ra, f0, f1)
                    fl[fl < 0] = np.nan
                    # get time
                    t = self.s.query(
                        "experiment == @e and condition == @c and strain == @s"
                    )["time"].to_numpy()
                    # store corrected fluorescence
                    bname = "c-" + f[0]
                    autofdict = {
                        "experiment": e,
                        "condition": c,
                        "strain": s,
                        "time": t,
                        bname: np.nanmean(fl, 1),
                        bname + " err": nanstdzeros2nan(fl, 1),
                    }
                    if useGPs:
                        flperod = samplewithGPs(
                            self,
                            t,
                            fl,
                            od,
                            flcvfn,
                            bd,
                            nosamples,
                            e,
                            c,
                            s,
                            figs,
                            **kwargs,
                        )
                    else:
                        # use only the replicates
                        flperod = fl / od
                        flperod[flperod < 0] = np.nan
                    # update dict
                    autofdict[bname + "perOD_err"] = naniqrzeros2nan(
                        flperod, 1
                    )
                    autofdict[bname + "perOD"] = np.nanmean(flperod, 1)
                    # add to data frames
                    addtodataframes(self, autofdict, bname)
                    addrecord(self, f[0], e, c, s)


def processref2(self, f, refstrain, figs, experiment, condition):
    """
    Process reference strain for spectral unmixing.

    Requires data with two fluorescence measurements.

    Use lowess to smooth the ratio of emitted fluorescence measurements
    so that the reference strain's data is corrected to zero as best
    as possible.

    Parameters
    ----------
    f: list of strings
        The fluorescence measurements. For example, ['GFP', 'AutoFL'].
    refstrain: string
        The reference strain. For example, 'WT'.
    figs: boolean
        If True, display fits of the fluorescence ratios.
    experiment: string
        The experiment to be corrected.
    condition: string
        The condition to be corrected.

    Returns
    -------
    qrfn: function
        The ratio of the two fluorescence values for the reference strain
        as a function of OD.
    """
    e, c = experiment, condition
    print(f"{e}: Processing reference strain {refstrain} for {f[0]} in {c}.")
    # refstrain data
    f0, f1, od = sunder.extractwells(
        self.r, self.s, e, c, refstrain, f + ["OD"]
    )
    if f0.size == 0 or f1.size == 0 or od.size == 0:
        raise errors.CorrectAuto(f"{e}: {refstrain} not found in {c}.")
    else:
        f0[f0 < 0] = np.nan
        f1[f1 < 0] = np.nan
        odf = od.flatten("F")
        odrefmean = np.mean(od, 1)
        qrf = (f1 / f0).flatten("F")
        if np.all(np.isnan(qrf)):
            raise errors.CorrectAuto(
                f"{e}: {refstrain} in {c} has too many NaNs."
            )
        # smooth to minimise autofluorescence in refstrain

        def choosefrac(frac):
            res = lowess(qrf, odf, frac)
            qrfn = interp1d(
                res[:, 0],
                res[:, 1],
                fill_value=(res[0, 1], res[-1, 1]),
                bounds_error=False,
            )
            flref = applyautoflcorrection(self, qrfn(odrefmean), f0, f1)
            return np.max(np.abs(flref))

        res = minimize_scalar(choosefrac, bounds=(0.1, 0.99), method="bounded")
        # calculate the relationship between qr and OD
        frac = res.x if res.success else 0.95
        res = lowess(qrf, odf, frac)
        qrfn = interp1d(
            res[:, 0],
            res[:, 1],
            fill_value=(res[0, 1], res[-1, 1]),
            bounds_error=False,
        )
        if figs:
            plt.figure()
            plt.plot(odf, qrf, ".", alpha=0.5)
            plt.plot(res[:, 0], res[:, 1])
            plt.xlabel("OD")
            plt.ylabel(f[1] + "/" + f[0])
            plt.title(e + ": " + refstrain + " in " + c)
            plt.show(block=False)
        # check autofluorescence correction for reference strain
        flref = applyautoflcorrection(self, qrfn(odrefmean), f0, f1)
        flrefperod = flref / od
        # set negative values to NaNs
        flref[flref < 0] = np.nan
        flrefperod[flrefperod < 0] = np.nan
        # store results
        bname = "c-" + f[0]
        autofdict = {
            "experiment": e,
            "condition": c,
            "strain": refstrain,
            "time": self.s.query(
                "experiment == @e and condition == @c and strain == @refstrain"
            )["time"].to_numpy(),
            bname: np.nanmean(flref, 1),
            bname + "perOD": np.nanmean(flrefperod, 1),
            bname + "_err": nanstdzeros2nan(flref, 1),
            bname + "perOD_err": nanstdzeros2nan(flrefperod, 1),
        }
        addtodataframes(self, autofdict, bname)
        return qrfn


def applyautoflcorrection(self, ra, f0data, f1data):
    """Correct for autofluorescence returning an array of replicates."""
    nr = f0data.shape[1]
    raa = np.reshape(np.tile(ra, nr), (np.size(ra), nr), order="F")
    return (raa * f0data - f1data) / (
        raa - self._gamma * np.ones(np.shape(raa))
    )


def addtodataframes(self, autofdict, bname):
    """Added dict of autofluorescence corrections to data frames."""
    autofdf = pd.DataFrame(autofdict)
    if bname not in self.s.columns:
        # extend dataframe
        self.s = pd.merge(self.s, autofdf, how="outer")
    else:
        # update dataframe
        self.s = gu.absorbdf(
            self.s,
            autofdf,
            ["experiment", "condition", "strain", "time"],
        )


def addrecord(self, fname, e, c, s):
    """Record correction."""
    self.sc.loc[
        (self.sc.experiment == e)
        & (self.sc.condition == c)
        & (self.sc.strain == s),
        fname + "_corrected_for_autofluorescence",
    ] = True


def nanstdzeros2nan(a, axis=None):
    """Like nanstd but setting zeros to nan."""
    err = np.nanstd(a, axis)
    err[err == 0] = np.nan
    return err


def naniqrzeros2nan(a, axis=None):
    """Interquartile range but setting zeros to nan."""
    iqr = np.nanquantile(a, 0.75, axis) - np.nanquantile(a, 0.25, axis)
    iqr[iqr == 0] = np.nan
    return iqr


def gethypers(self, exp, con, s, dtype="gr"):
    """Find parameters for GP from sc data frame."""
    sdf = self.sc[
        (self.sc.experiment == exp)
        & (self.sc.condition == con)
        & (self.sc.strain == s)
    ]
    try:
        cvfn = sdf["gp_for_gr"].values[0]
        hypers = [
            sdf[col].values[0] for col in sorted(sdf.columns) if "hyper" in col
        ]
        if np.any(np.isnan(hypers)):
            return None, None
        else:
            return hypers, cvfn
    except KeyError:
        return None, None


def samplewithGPs(
    self, t, fl, od, flcvfn, bd, nosamples, e, c, s, figs, **kwargs
):
    """Generate extra samples of flperod using Gaussian processes."""
    hypers, cvfn = gethypers(self, e, c, s)
    if hypers is None or cvfn is None:
        raise SystemExit(
            f"\nYou first must run getstats for {s} in {c} for {e}."
        )
    # initialise GP for log ODs
    t1 = np.tile(t, od.shape[1])
    od1 = np.reshape(od, od.size, order="F")
    od1 = od1[np.argsort(t1)]
    t1 = np.sort(t1)
    # bounds are irrelevant because parameters are optimal
    go = getattr(gp, cvfn + "GP")(
        {0: (-5, 5), 1: (-4, 4), 2: (-5, 2)},
        t1,
        np.log(od1),
    )
    go.lth_opt = hypers
    go.predict(t)
    # run GP for log fluorescence
    print(f"Fitting fluorescence for {e}: {s} in {c}")
    ff = fitderiv(
        t,
        fl,
        cvfn=flcvfn,
        bd=bd,
        logs=True,
        stats=False,
        **kwargs,
    )
    if ff.success:
        print()
        if figs:
            plt.figure()
            ff.plotfit(
                ylabel="log fluorescence",
                figtitle=f"{e}: {s} in {c}",
            )
            plt.show(block=False)
    else:
        print(f"\n-> Fitting fluorescence failed for {e}: {s} in {c}.\n")
        return np.nan * np.ones(fl.shape)
    # sample
    lod_samples = go.sample(nosamples)
    lfl_samples = ff.fitderivsample(nosamples)[0]
    flperod = np.exp(lfl_samples - lod_samples)
    return flperod



# functions to plot from the data frames
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

import om_code.omerrors as errors
import om_code.omgenutils as gu


def plotplate(basedf, exps, dtype):
    """
    Plot the data for each well following the layout of a 96-well plate.

    Parameters
    ----------
    basedf: DataFrame
        The r dataframe.
    exps: float
        The name of the experiments.
    dtype: float
        The data type to be plotted: 'OD', 'GFP', etc.
    """
    for e in exps:
        plt.figure()
        # first create an empty plate - in case of missing wells
        ax = []
        for rowl in range(8):
            for coll in np.arange(1, 13):
                sindex = coll + 12 * rowl
                axi = plt.subplot(8, 12, sindex)
                ax.append(axi)
                plt.tick_params(labelbottom=False, labelleft=False)
                # label well locations
                for j in range(12):
                    if sindex == j + 1:
                        plt.title(j + 1)
                for j, k in enumerate(np.arange(1, 96, 12)):
                    if sindex == k:
                        plt.ylabel("ABCDEFGH"[j] + " ", rotation=0)
        # fill in the wells that have been measured
        for pl in basedf.query("experiment == @e")["well"].unique():
            rowl = "ABCDEFGH".index(pl[0])
            coll = int(pl[1:])
            sindex = coll + 12 * rowl
            wd = basedf.query("experiment == @e and well == @pl")
            ax[sindex - 1].plot(
                wd["time"].to_numpy(), wd[dtype].to_numpy(), "-"
            )
        plt.suptitle(e + ": " + dtype)
        plt.show(block=False)


def plot_wells(
    x,
    y,
    basedf,
    exps,
    cons,
    strs,
    style="condition",
    size=None,
    kind="line",
    col=None,
    row=None,
    xlim=None,
    ylim=None,
    title=None,
    figsize=None,
    messages=False,
    **kwargs,
):
    """
    Plot data from the individual wells.

    Data for each experiment, condition, and strain are plotted in
    a separate figure unless row and col are specified.
    """
    for e in exps:
        if row and col:
            # use facetgrid to show multiple plots simultaneously
            df = basedf.query(
                "experiment == @e and condition == @cons and strain == @strs"
            )
            sfig = sns.FacetGrid(df, row=row, col=col)
            for (row_var, col_var), facet_df in df.groupby([row, col]):
                ax = sfig.axes[
                    sfig.row_names.index(row_var),
                    sfig.col_names.index(col_var),
                ]
                sns.lineplot(x=x, y=y, hue="well", data=facet_df, ax=ax)
                ax.set(xlabel="", ylabel="")
            sfig.set_axis_labels(x, y)
            sfig.set_titles()
            if title:
                sfig.fig.suptitle(title)
            else:
                sfig.fig.suptitle(e)
            if xlim is not None:
                plt.xlim(xlim)
            if ylim is not None:
                plt.ylim(ylim)
            if figsize and len(figsize) == 2:
                sfig.fig.set_figwidth(figsize[0])
                sfig.fig.set_figheight(figsize[1])
            plt.tight_layout()
            plt.show(block=False)
        else:
            # create one plot for each strain and condition
            for c in cons:
                for s in strs:
                    df = basedf.query(
                        "experiment == @e and condition == @c and strain == @s"
                    )
                    if df.empty:
                        if messages:
                            print(e + ":", "No data found for", s, "in", c)
                    else:
                        sfig = sns.relplot(
                            x=x,
                            y=y,
                            data=df,
                            hue="well",
                            kind=kind,
                            style=style,
                            size=size,
                            **kwargs,
                        )
                        if title:
                            sfig.fig.suptitle(title)
                        else:
                            sfig.fig.suptitle(e + ": " + s + " in " + c)
                        if xlim is not None:
                            plt.xlim(xlim)
                        if ylim is not None:
                            plt.ylim(ylim)
                        plt.tight_layout()
                        plt.show(block=False)


def plot_rs(
    x,
    y,
    basedf,
    exps,
    cons,
    strs,
    hue="strain",
    style="condition",
    size=None,
    kind="line",
    col=None,
    row=None,
    height=5,
    aspect=1,
    xlim=None,
    ylim=None,
    title=None,
    figsize=None,
    sortby=False,
    returnfacetgrid=False,
    **kwargs,
):
    """Plot time-series data from the .r or .s dataframes."""
    # plot time series
    df = basedf.query(
        "experiment == @exps and condition == @cons and strain == @strs"
    )
    if df.empty:
        print("No data found")
    else:
        if sortby:
            df = df.sort_values(by=gu.makelist(sortby))
        # add warnings for poor choice of seaborn's parameters - may cause
        # inadvertent averaging
        if hue == style:
            print(
                'Warning: "hue" and "style" have both been set to "'
                + hue
                + '" and there may be unintended averaging'
            )
        if (
            x != "commontime"
            and len(df["experiment"].unique()) > 1
            and hue != "experiment"
            and size != "experiment"
            and style != "experiment"
            and col != "experiment"
        ):
            print(
                "Warning: there are multiple experiments, but neither "
                '"hue", "style", nor "size" is set to "experiment" and there'
                " may be averaging over experiments"
            )

        if "units" not in kwargs:
            # try to augment df to allow seaborn to estimate errors
            df = augmentdf(df, y)
        # plot
        sfig = sns.relplot(
            x=x,
            y=y,
            data=df,
            hue=hue,
            kind=kind,
            style=style,
            errorbar="sd",
            size=size,
            col=col,
            row=row,
            aspect=aspect,
            height=height,
            **kwargs,
        )
        sfig.set_xlabels(gu.rm_under(x))
        sfig.set_ylabels(gu.rm_under(y))
        if title:
            sfig.fig.suptitle(title)
        if xlim is not None:
            sfig.set(xlim=xlim)
        if ylim is not None:
            sfig.set(ylim=ylim)
        if figsize and len(figsize) == 2:
            sfig.fig.set_figwidth(figsize[0])
            sfig.fig.set_figheight(figsize[1])
        plt.show(block=False)
        if returnfacetgrid:
            return sfig
        else:
            return None


def plot_sc(
    x,
    y,
    basedf,
    exps,
    cons,
    strs,
    hue="strain",
    style="condition",
    size=None,
    kind="line",
    col=None,
    row=None,
    height=5,
    aspect=1,
    xlim=None,
    ylim=None,
    figsize=None,
    title=None,
    sortby=False,
    **kwargs,
):
    """Plot summary statistics from the .sc dataframe."""
    # plot summary stats
    df = basedf.query(
        "experiment == @exps and condition == @cons and strain == @strs"
    )
    xcols = df.columns[df.columns.str.startswith(x)]
    ycols = df.columns[df.columns.str.startswith(y)]
    df = df[
        np.unique(
            ["experiment", "condition", "strain"] + list(xcols) + list(ycols)
        )
    ].dropna()
    if df.empty:
        print("No data found")
    else:
        if sortby:
            df = df.sort_values(by=gu.makelist(sortby))
        sfig = sns.relplot(
            x=x,
            y=y,
            data=df,
            hue=hue,
            kind="scatter",
            style=style,
            size=size,
            col=col,
            row=row,
            aspect=aspect,
            height=height,
            **kwargs,
        )
        sfig.set_xlabels(gu.rm_under(x))
        sfig.set_ylabels(gu.rm_under(y))
        if xlim is not None:
            sfig.set(xlim=xlim)
        if ylim is not None:
            sfig.set(ylim=ylim)
        if row is None and col is None:
            # add error bars
            # find coordinates of points in relplot
            xc, yc = [], []
            for point_pair in sfig.ax.collections:
                for xp, yp in point_pair.get_offsets():
                    xc.append(xp)
                    yc.append(yp)
            # add error bars
            xerr = df[x + "_err"] if x + "_err" in df.columns else None
            yerr = df[y + "_err"] if y + "_err" in df.columns else None
            sfig.ax.errorbar(
                xc,
                yc,
                xerr=xerr,
                yerr=yerr,
                fmt=" ",
                ecolor="dimgray",
                alpha=0.5,
            )
        plt.show(block=False)


def plotfinddf(self, x, y):
    """
    Find the correct dataframe for plotting y versus x.

    Parameters
    ----------
    self: a platereader instance
    x: string
        Name of x-variable.
    y: string
        Name of y-variable.

    Returns
    -------
    basedf: dataframe
        The dataframe that contains the x and y variables.
    dfname: string
        The name of the dataframe.
    """
    # choose the correct dataframe
    if hasattr(self, "r") and x in self.r.columns and y in self.r.columns:
        # raw data (with wells)
        basedf = self.r
        dfname = "r"
    elif x in self.s.columns and y in self.s.columns:
        # processed data (no wells)
        basedf = self.s
        dfname = "s"
    elif x in self.sc.columns and y in self.sc.columns:
        # summary stats
        basedf = self.sc
        dfname = "sc"
    else:
        raise errors.PlotError(
            f"The variables x= {x}"
            + f" and y= {y}"
            + " cannot be plotted against each other because they are not in "
            + " the same dataframe"
        )
    return basedf, dfname


def augmentdf(df, datatype):
    """
    Augment dataframe to allow Seaborn to errors.

    Use 'err' (if present in the dataframe) to allow Seaborn to generate
    errors in relplot, otherwise returns original dataframe.

    Note we call seaborn with errorbar = "sd" and so use sqrt(3/2) * error
    because seaborn calculates the standard deviation from the augmented data
    (the mean, the mean + std, and the mean - std) and so gets
    std/sqrt(3/2) otherwise because there are three data points.
    """
    if datatype + "_err" in df:
        derr = datatype + "_err"
    elif "mean" in datatype and datatype.split("_mean")[0] + "_err" in df:
        derr = datatype.split("_mean")[0] + "_err"
    else:
        derr = False
        # returned if df is df_r
        return df
    if derr:
        df.insert(0, "augtype", "mean")
        mn = df[datatype].to_numpy()
        err = df[derr].to_numpy()
        # add std
        dfp = df.copy()
        dfp[datatype] = mn + np.sqrt(3 / 2) * err
        dfp["augtype"] = "+err"
        # minus std
        dfm = df.copy()
        dfm[datatype] = mn - np.sqrt(3 / 2) * err
        dfm["augtype"] = "-err"
        # concat
        df = pd.concat([df, dfp, dfm], ignore_index=True)
    return df



import numpy as np


def analyseTecan(dfd, rcontents, experiment):
    """
    Parse an imported Excel file from a Tecan F200 or M200 plate reader.

    Parameters
    ----------
    dfd: dataframe
        Created by importing the data from a file using Panda's read_excel.
    rcontents: dataframe
        Created by analyseContentsofWells.
    experiment: string
        The name of the experiment.

    Returns
    ------
    rdict: list of dictionaries
        Describes the contents of the plate by experiment, condition, strain,
        time, and well.
    datatypes: list of strings
        Delineates all the types of data in the experiment and is minimally
        ['OD'].
    """
    # extract datatypes
    datatypes = (
        dfd[dfd.columns[0]]
        .iloc[
            np.nonzero(
                dfd[dfd.columns[0]]
                .str.startswith("Cycle Nr", na=False)
                .to_numpy()
            )[0]
            - 1
        ]
        .to_numpy()
    )
    # if only OD data measured
    if not isinstance(datatypes[0], str):
        datatypes = ["OD"]
    # extract times of measurements
    t = (
        dfd.loc[
            dfd[dfd.columns[0]].str.startswith("Time [s]", na=False),
            dfd.columns[1] :,
        ]
        .dropna(axis="columns")
        .mean()
        .to_numpy()
        .astype("float")
        / 3600.0
    )
    # deal with overflows
    df = dfd.replace("OVER", np.nan)
    cols = df.columns
    # extract data
    # add to dataframe
    df.index = df[cols[0]]
    rdict = []
    for x in np.arange(1, 13):
        for y in "ABCDEFGH":
            well = y + str(x)
            if well in df.index:
                data = df.loc[well, cols[1] :].to_numpy(dtype="float")
                if data.ndim == 1:
                    data = data[None, :]
                if (
                    rcontents[well][0] is not None
                    and rcontents[well][1] is not None
                ):
                    for j in range(len(t)):
                        cons = {
                            "experiment": experiment,
                            "condition": rcontents[well][0],
                            "strain": rcontents[well][1],
                            "time": t[j],
                            "well": well,
                        }
                        dats = {
                            datatype: data[i, j]
                            for i, datatype in enumerate(datatypes)
                        }
                        cons.update(dats)
                        rdict.append(cons)
    return rdict, datatypes



import numpy as np
from scipy import integrate
from scipy.interpolate import interp1d
import matplotlib.pylab as plt
import om_code.omerrors as errors
from om_code.stats import statserr


def getfitnesspenalty(
    self, ref, com, y="gr", abs=False, figs=True, nosamples=100, norm=False
):
    """
    Estimate the difference in fitness between two strains.

    Calculate the area between typically two growth rate versus OD
    curves, normalised by the length along the OD-axis where they overlap.

    Parameters
    -----------
    ref: list of strings
        For only a single experiment, a list of two strings. The first string
        specifies the condition and the second specifies the strain to be used
        for the reference to which fitness is to be calculated.
        With multiple experiments, a list of three strings. The first string
        specifies the experiment, the second specifies the condition, and the
        third specifies the strain.
    com: list of strings
        For only a single experiment, a list of two strings. The first string
        specifies the condition and the second specifies the strain to be
        compared with the reference.
        With multiple experiments, a list of three strings. The first string
        specifies the experiment, the second specifies the condition, and the
        third specifies the strain.
    y: string, optional
        The variable to be compared.
    figs: boolean, optional
        If True, a plot of the area between the two growth rate versus OD
        curves is shown.
    nosamples: integer
        The number bootstraps used to estimate the error.
    norm: boolean
        If True, returns the mean and variance of the area under the reference
        strain for normalisation.

    Returns
    -------
    fp: float
        The area between the two curves.
    err: float
        An estimate of the error in the calculated error, found by
        bootstrapping.
    reffp: float, optional
        The area beneath the reference strain.
    referr: float, optional
        An estimate of the erroe in the calculated area for the reference
        strain.

    Example
    -------
    >>> p.getfitnesspenalty(['1% raf 0.0µg/ml cyclohex', 'WT'],
    ...                     ['1% raf 0.5µg/ml cyclohex', 'WT'])
    """
    if len(self.allexperiments) == 1:
        ref.insert(0, self.allexperiments[0])
        com.insert(0, self.allexperiments[0])
    # get and sample from Gaussian processes
    if nosamples and y == "gr":
        # estimate errors
        try:
            # sample from Gaussian process
            f0s, g0s, h0s = self.progress["getstatsGP"][ref[0]][ref[1]][
                ref[2]
            ].fitderivsample(nosamples)
            f1s, g1s, h1s = self.progress["getstatsGP"][com[0]][com[1]][
                com[2]
            ].fitderivsample(nosamples)
            xsref, ysref = np.exp(f0s), g0s
            xscom, yscom = np.exp(f1s), g1s
        except KeyError:
            raise errors.GetFitnessPenalty(
                "getstats('OD') needs to be run for these strains to "
                "estimate errors or else set nosamples= 0"
            )
    else:
        # no estimates of errors
        if nosamples:
            print(
                "Cannot estimate errors - require y= 'gr' and a recently "
                "run getstats"
            )
        xsref = self.s.query(
            "experiment == @ref[0] and condition == @ref[1] and "
            "strain == @ref[2]"
        )["OD mean"][:, None]
        ysref = self.s.query(
            "experiment == @ref[0] and condition == @ref[1] and "
            "strain == @ref[2]"
        )[y].to_numpy()[:, None]
        xscom = self.s.query(
            "experiment == @com[0] and condition == @com[1] and "
            "strain == @com[2]"
        )["OD mean"].to_numpy()[:, None]
        yscom = self.s.query(
            "experiment == @com[0] and condition == @com[1] and "
            "strain == @com[2]"
        )[y].to_numpy()[:, None]
        if xsref.size == 0 or ysref.size == 0:
            print(ref[0] + ": Data missing for", ref[2], "in", ref[1])
            return np.nan, np.nan
        elif xscom.size == 0 or yscom.size == 0:
            print(com[0] + ": Data missing for", com[2], "in", com[1])
            return np.nan, np.nan
    fps = np.zeros(xsref.shape[1])
    nrm = np.zeros(xsref.shape[1])
    samples = zip(
        np.transpose(xsref),
        np.transpose(ysref),
        np.transpose(xscom),
        np.transpose(yscom),
    )
    # process samples
    for j, (xref, yref, xcom, ycom) in enumerate(samples):
        # remove any double values in OD because of OD plateau'ing
        uxref, uiref = np.unique(xref, return_inverse=True)
        uyref = np.array(
            [
                np.median(yref[np.nonzero(uiref == i)[0]])
                for i in range(len(uxref))
            ]
        )
        uxcom, uicom = np.unique(xcom, return_inverse=True)
        uycom = np.array(
            [
                np.median(ycom[np.nonzero(uicom == i)[0]])
                for i in range(len(uxcom))
            ]
        )
        # interpolate data
        iref = interp1d(uxref, uyref, fill_value="extrapolate", kind="slinear")
        icom = interp1d(uxcom, uycom, fill_value="extrapolate", kind="slinear")
        # find common range of x
        uxi = np.max([uxref[0], uxcom[0]])
        uxf = np.min([uxref[-1], uxcom[-1]])
        # perform integration to find normalized area between curves
        if abs:

            def igrand(x):
                return np.abs(iref(x) - icom(x))

        else:

            def igrand(x):
                return iref(x) - icom(x)

        fps[j] = integrate.quad(igrand, uxi, uxf, limit=100, full_output=1)[
            0
        ] / (uxf - uxi)
        if norm:
            # calculate area under curve of reference strain as a normalisation
            def igrand(x):
                return iref(x)

            nrm[j] = integrate.quad(
                igrand, uxi, uxf, limit=100, full_output=1
            )[0] / (uxf - uxi)
        # an example figure
        if figs and j == 0:
            plt.figure()
            plt.plot(uxref, uyref, "k-", uxcom, uycom, "b-")
            x = np.linspace(uxi, uxf, np.max([len(uxref), len(uxcom)]))
            plt.fill_between(x, iref(x), icom(x), facecolor="red", alpha=0.5)
            plt.xlabel("OD")
            plt.ylabel(y)
            plt.legend(
                [
                    ref[0] + ": " + ref[2] + " in " + ref[1],
                    com[0] + ": " + com[2] + " in " + com[1],
                ],
                loc="upper left",
                bbox_to_anchor=(-0.05, 1.15),
            )
            plt.show(block=False)
    if norm:
        return (
            np.median(fps),
            statserr(fps),
            np.median(nrm),
            statserr(nrm),
        )
    else:
        return np.median(fps), statserr(fps)



import datetime
import logging
from io import StringIO
import functools


def initlog(version=None):
    """Enable logging."""
    starttime = "{:%Y-%b-%d %H:%M:%S}".format(datetime.datetime.now())
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)
    logstream = StringIO()
    loghandler = logging.StreamHandler(logstream)
    loghandler.setLevel(logging.INFO)
    logger.addHandler(loghandler)
    logger.propagate = False
    # start log
    if version:
        logger.info("version " + f"{version}")
    logger.info(f"{starttime}\n")
    return logger, logstream

    ###


def log(func):
    """To decorate functions whose output should be logged."""

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        res = func(self, *args, **kwargs)
        # store call to function in logstring
        methodname = func.__name__
        logstring = methodname + "("
        if any(args):
            argslist = [repr(a) for a in args]
        else:
            argslist = []
        if any(kwargs):
            kwargslist = [f"{k}={v!r}" for k, v in kwargs.items()]
        else:
            kwargslist = []
        logstring += ", ".join(argslist + kwargslist) + ")\n"
        self.logger.info(logstring)
        return res

    return wrapper
